//------------------------------------------------------------------
//
// Copyright (c) 1999 -2009 Tapko Technologies GmbH
//
// PROPRIETARY RIGHTS of Tapko Technologies GmbH are involved
// in the subject matter of this material. All manufacturing,
// reproduction, use, and sales rights pertaining to this subject matter
// are governed by the license agreement. The recipient of this software
// implicitly accepts the terms of the license.
//
// This file is part of the Libraries of Tapko Technologies GmbH
// (e.g. TAPKO EIB/KNX Communication Stack)
//
//------------------------------------------------------------------
// $Workfile: TaUtil_Timer.c $
// $Archive: /eib_stk/appl/A1008TA_KIMaip_Eval/src/TaUtil_Timer.c $
//
// Project:           EIB-Communication-Stack
// 
// Description:
// The timer system is based on timer variables, which are updated in interrupt
// The check of the timer (calculation of time), is done in a cyclic part.
// The interrupt part in in the specific project.
//
// Compile-switches:
//
// $Author: Thiere_m $         
// 
// $Date: 20.05.11 10:53 $
// $Revision: 33 $
//
//------------------------------------------------------------------
// 09.07.2003 TM_GENERATE_LL_TIMEOUT introduced
// 21.03.2005 TM_Cycle: time of disabling interrupt reduced
// 08.09.2009 TM_StopTimer checks parameter
//------------------------------------------------------------------
#include "project.h"
#include "eib_stk.h"

#include "TaUtil_types.h"
#include "TaUtil_Timer.h"

#ifdef DEBUG_WIN32
  #include "time.h"
#endif
#ifdef KNX_STACK
#ifdef TM_GENERATE_LL_TIMEOUT
#include <../ll/tpuart/tpuart.h>
#endif
#endif // KNX_STACK
//------------------------------------------------------------------
// 100 ms timer tick
//------------------------------------------------------------------
#ifdef TM_100MS_TICK
unsigned short TM_100msLast;
TA_BOOL TM_100msTick;
#endif

//------------------------------------------------------------------
unsigned short TM_GetTime(unsigned char mode)
{
  switch (mode & TM_MODE_TIME_MASK)
  {
    case TM_MODE_1MS: 
      return (unsigned short) (TMB_GetTime(0) & 0x7fff);

    case TM_MODE_1S:
      return (unsigned short) (TMB_GetTime(TM_MODE_1S) & 0x7fff);

    case TM_MODE_1MIN:
      return (unsigned short) ((TMB_GetTime(TM_MODE_1S) / 60L) & 0x7fff);

    case TM_MODE_1H:
      return (unsigned short) ((TMB_GetTime(TM_MODE_1S) / 3600L) & 0x7fff);
  }
  return 0;
}

//------------------------------------------------------------------
// Initialization of the timer
//------------------------------------------------------------------
void TM_Init(void)
{
  unsigned char count = 0;

  _DINT();
  for (count = 0; count < TM_timerNr; count++)  // loop over all timer
    TM_tmData[count].state = 0;                // stop timer
#ifdef TM_100MS_TICK
  TM_100msLast = TM_GetTime(0) & 0x7fff;
  TM_100msTick = TA_FALSE;
#endif
  _EINT();
}

//------------------------------------------------------------------
// Stops one timer
// input:     timerNr
//------------------------------------------------------------------
void TM_StopTimer(unsigned char timerNr)
{
  if (timerNr >= TM_timerNr)                     // check for valid timerNr
    return;

  TM_tmData[timerNr].state = 0;                  // stop timer
}


//------------------------------------------------------------------
// Start one timer
// input: timerNr
//        time in ms / s
//        mode (cyclic, ms / s)
// for microcontroller: interrupts are disabled (monitor)
//------------------------------------------------------------------
void TM_StartTimer(unsigned char timerNr, unsigned short time, unsigned char mode)
{
  TM_tmStruct *timerStruct;                      // pointer to timer struct

  if (timerNr >= TM_timerNr)                     // check for valid timerNr
    return;

  timerStruct = &TM_tmData[timerNr];             // store pointer to current timer struct

  timerStruct->startTime = TM_GetTime(mode) & 0x7fff;
  timerStruct->delayTime = time;                 // store delay time
  timerStruct->mode = mode;
  timerStruct->state = TM_STATE_RUNNING;         // set state to running
}

//------------------------------------------------------------------
// Returns the state of one timer
// If the timer was elapsed this function returns _once_ elapsed.
// input: timerNr
//
// return value: TA_TRUE: timer elapsed
//               TA_FALSE: timer running or stopped
//------------------------------------------------------------------
TA_BOOL TM_GetState(unsigned char timerNr)
{
  TM_tmStruct *timerStruct;                      // pointer to timer structure

  if (timerNr >= TM_timerNr)                     // check for valid timerNr
    return TA_FALSE;

  timerStruct = &TM_tmData[timerNr];             // store pointer to current timer struct
                                                 // check if timer was elapsed
  if ((timerStruct->state & TM_STATE_ELAPSED) != 0)
  {
    timerStruct->state &= ~TM_STATE_ELAPSED;     // restore flag that timer was elapsed
    return TA_TRUE;
  }
  return TA_FALSE;
}

//------------------------------------------------------------------
// cyclic part of the timer
// this part calculates and updates all the timer (structures).
// the basis for the calculation are the timer variables, which are updated
// in interrupt
//------------------------------------------------------------------
TA_BOOL TM_Cycle(void)
{
  unsigned short tmValue;                        // variable to store the current timer counter
  unsigned short l_startTime;
  unsigned short l_delayTime;
  unsigned char count;                           // counter over all the timer
  TM_tmStruct *timerStruct;                      // pointer to timer structure
  unsigned char globalState = 0;

  SC_SetTaskActive(SC_TASK_TM, TA_FALSE);
#ifdef TM_100MS_TICK
  tmValue = TM_GetTime(0);
  TM_100msTick = (((TM_100msLast + 100  <= tmValue && TM_100msLast <= tmValue) ||
                 (TM_100msLast + 100  <= tmValue + 0x8000 && TM_100msLast > tmValue)));
  if (TM_100msTick)
    TM_100msLast = (TM_100msLast + 100) & 0x7fff;
#endif

  for (count = 0; count < TM_timerNr; count++)   // loop over all timer
  {
    timerStruct = &TM_tmData[count];             // store pointer to timer struct

                                                 // is timer running
    if ((timerStruct->state & TM_STATE_RUNNING) != 0)
    {
      _DINT();
      l_startTime = timerStruct->startTime;
      l_delayTime = timerStruct->delayTime;
      _EINT();

      tmValue = TM_GetTime(timerStruct->mode);   // store current timer counter (s / ms)

                                                 // calculate if timer is elapsed
      if (l_delayTime == 0 ||
          ((l_startTime + l_delayTime  <= tmValue && l_startTime <= tmValue) ||
          (l_startTime + l_delayTime  <= tmValue + 0x8000 && l_startTime > tmValue)))
      {
        _DINT();
        if (l_startTime == timerStruct->startTime)
        {
          if ((timerStruct->mode & TM_MODE_CYCLE) == 0)
          {                                      // not cyclic => timer elapsed, stopped
            timerStruct->state = TM_STATE_ELAPSED;
          }
          else
          {                                      // cyclic => timer elapsed
            timerStruct->state |= TM_STATE_ELAPSED;
                                                 // store new start time
            timerStruct->startTime = (l_startTime + l_delayTime) & 0x7fff;
          }
        }
        _EINT();
      }
      globalState |= timerStruct->state;
    }

                                                 // check for special timer features
    if ((timerStruct->state & TM_STATE_ELAPSED) != 0)
    {
#ifdef KNX_STACK
#ifdef TM_GENERATE_LL_TIMEOUT
      if ((timerStruct->mode & TM_MODE_LL_EVENT) != 0)
      {                                          // generate event for LL
        _DINT();
        timerStruct->state = 0;
        TUI_RcvTimeout();
        _EINT();
      }
#endif
#endif // KNX_STACK
    }
  }
  return (globalState & TM_STATE_ELAPSED) == TM_STATE_ELAPSED;
}

//------------------------------------------------------------------
//------------------------------------------------------------------
void TM_Delay(unsigned short time)
{
  #ifndef DEBUG_WIN32

    volatile unsigned short tmAktuell;
    unsigned short tmStart;

    tmStart= TM_GetTime(0);

    do
    {
      tmAktuell= TM_GetTime(0);
    } while (!(((tmStart + time <= tmAktuell) && (tmStart <= tmAktuell)) ||
      ((tmStart + time <= tmAktuell + 0x8000) && (tmStart > tmAktuell))));
  #else

    clock_t goal;
    goal = time + clock();
    while( goal > clock() )
      ;

  #endif
}


//------------------------------------------------------------------
//------------------------------------------------------------------
TA_BOOL TM_GetInfo(unsigned char timerNr, unsigned short *remainingTime)
{
  TM_tmStruct *timerStruct;                      // pointer to timer structure
  unsigned short tmValue;

  *remainingTime=0;

  if (timerNr >= TM_timerNr)                     // check for valid timerNr
    return TA_FALSE;

  timerStruct = &TM_tmData[timerNr];             // store pointer to current timer struct
                                                 // check if timer is running
  if ((timerStruct->state & TM_STATE_RUNNING) != 0)
  {
    tmValue = TM_GetTime(timerStruct->mode);     // store current timer counter (s / ms)

    if (timerStruct->startTime > tmValue)
      tmValue += 0x8000;

    *remainingTime = timerStruct->startTime + timerStruct->delayTime - tmValue;

    return TA_TRUE;
  }
  return TA_FALSE;
}

#pragma vector=TIMERA1_VECTOR
__interrupt void ISR_Timer(void)           // Interrupt-routine for TimerA
{
  switch (TAIV)
  {
      
    case TAIV_TACCR2:          // Capture / Compare 2 (1 ms Timer)
      TACCR2 += TM_CLOCK_COUNT;

      TM_counter1ms += TM_MS_TIMER_ADDER;
      TM_counter1s_help++;
      if (TM_counter1s_help >= TM_S_TIMER_ADDER)
      {
        TM_counter1s_help = 0;
        TM_counter1s++;
      }
      SC_SetTaskActive(SC_TASK_TM, TA_TRUE);
      break;
    
    case TAIV_TAIFG:                           // Timer overflow
      break;
  }
}

